#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char *codesegs[] = {"p1p12p3pX4p5p5pZ7pi", // Init code
		">_#\"0g91+8*1g*+4g+g:8p5g!*", // Instruction fetch
		"Y6p> #p:#66#+g#2g#g-#6_$", // Handler lookup
		"6g1+g6pW>1#p-#+:#g:#7:#\\6#gg#+\\_g7gp$$$", // Handler copy
		"h0g2g+91+8*%0p1g3g+55*%1p0#\"<", // Handler marker+return code
		0};

// Replace X with data (befunge program) start adr
// Replace Y with instruction lookup table
// Replace Z with adr of h
// Replace W with width of h
// Replace h with gap for handler to be copied to
// Replace i with code to fill in handler table - Y1+[:number\p2+]$

char *handlers[][2] = {{"0123456789!@$&*\\+-:~`./, ","8g7g7+p"},
		{"<","01-2p03p"},
		{"_","!2*1-2p03p"},
		{">","12p03p"},
		{"g","91+8**+4g+g"},
		{"p","91+8**+4g+p"},
		{"v","02p13p"},
		{"|","02p!2*1-3p"},
		{"^","02p01-3p"},
		{"#","0g2g+0p1g3g+1p"},
		{"\"","5g!5p"},
		{"?","01-1>\\# ?23>\\# ?0\\pp$"},
		{"","8g:84*2+-!:!5p>0$_"}};

#define NUM_HANDLERS 13

// Size to treat numbers as
#define NUMBER_WIDTH 12

#define TABLE_SIZE (table_entries*2)
#define HANDLER_SIZE (NUM_HANDLERS*handler_width)
#define TABLESETUP_SIZE (NUMBER_WIDTH+3+(NUMBER_WIDTH+5)*table_entries)
#define APPROX_LEN (codelen+TABLESETUP_SIZE+(numvars*NUMBER_WIDTH)+handler_width+TABLE_SIZE+HANDLER_SIZE)

int handler_width,table_entries;
int codelen,numvars;
int running_total;

void output_num_recursive(int n,int *charcount)
{
	if (n == 0)
		return;
	if (n < 10)
	{
		fputc(n+'0',stdout);
		(*charcount)++;
		return;
	}
	// .. else > 10...
	if ((int) (n/9) > 1)
	{
		output_num_recursive(n/9,charcount); // Output number /9
		printf("9*"); // Times it by nine
		(*charcount)+=2;
	}
	else
	{
		printf("9");
		(*charcount)++;
	}
	// Now output remainder
	if (n%9 != 0)
	{
		printf("%d+",n%9);
		(*charcount)+=2;
	}
}

void output_num(int n)
{
	int charcount;
	// Output code to push 'n' onto the stack
	// Avoid stringmode, and output length must be NUMBER_WIDTH
	charcount = 0;
	output_num_recursive(n,&charcount);
	if (charcount > NUMBER_WIDTH)
	{
		fprintf(stderr,"\n*** Number %d too wide ***\n",n);
		exit(1);
	}
	while (charcount < NUMBER_WIDTH)
	{
		putc(' ',stdout);
		charcount++;
	}
	running_total += NUMBER_WIDTH;
}

void output_x()
{
	// output 'x', aka data start address
	output_num(APPROX_LEN);
}

void output_y()
{
	// output 'y', aka instruction table address
	output_num(codelen+TABLESETUP_SIZE+(numvars*NUMBER_WIDTH)+handler_width);
}

void output_z()
{
	// output 'z', aka address of 'h'
	int tmp,tmp2,count;
	tmp=0;
	count=0;
	while (codesegs[tmp])
	{
		tmp2=0;
		while (codesegs[tmp][tmp2])
		{
			if (codesegs[tmp][tmp2] == 'X')
				count+=NUMBER_WIDTH;
			else if (codesegs[tmp][tmp2] == 'Y')
				count+=NUMBER_WIDTH;
			else if (codesegs[tmp][tmp2] == 'Z')
				count+=NUMBER_WIDTH;
			else if (codesegs[tmp][tmp2] == 'W')
				count+=NUMBER_WIDTH;
			else if (codesegs[tmp][tmp2] == 'h')
			{
				output_num(count); // Found where it should be
				return;
			}
			else if (codesegs[tmp][tmp2] == 'i')
				count+=TABLESETUP_SIZE;
			else
				count++;
			tmp2++;
		}
		tmp++;
	}
	// Doh!
	fprintf(stderr,"\n*** output_z finished! ***\n");
	exit(1);
}

void output_w()
{
	// output 'w', aka handler width
	output_num(handler_width);
}

void output_h()
{
	// output 'h', aka gap for handler itself
	int tmp;
	for (tmp=0;tmp<handler_width;tmp++)
		putc(' ',stdout);
	running_total += handler_width;
}

int handler_for_instr(int i)
{
	// Find the handler for instruction number i
	int tmp;
	tmp=0;
	while (tmp < NUM_HANDLERS)
	{
		i-=strlen(handlers[tmp][0]);
		if (i < 0)
			return tmp;
		tmp++;
	}
	if (i == 0)
		return tmp-1; // String
	// doh!
	fprintf(stderr,"\n*** handler_for_instr finished! ***\n");
	exit(1);
}

void output_i()
{
	int tmp;
	// output 'i', aka code to fill handler table
	output_y(); // output table start
	printf("1+"); // skip 1st instr name
	for (tmp=0;tmp<table_entries;tmp++)
	{
		printf(":");
		output_num(codelen+TABLESETUP_SIZE+(numvars*NUMBER_WIDTH)+handler_width+TABLE_SIZE+(handler_for_instr(tmp)*handler_width)); // Output address of handler
		printf("\\p2+");
		running_total += 5;
	}
	printf("$");
	running_total += 3;
}

void output_table()
{
	// Output empty handler lookup table
	int tmp,tmp2;
	tmp=0;
	while (tmp < NUM_HANDLERS)
	{
		tmp2=0;
		while (handlers[tmp][0][tmp2])
		{
			printf("%c ",handlers[tmp][0][tmp2]); // Output char
			running_total += 2;
			tmp2++;
		}
		tmp++;
	}
	// ... and stringmode
	printf("%c ",0);
	running_total += 2;
}

void output_handlers()
{
	// Output the handler code
	int tmp,tmp2;
	tmp=0;
	while (tmp < NUM_HANDLERS)
	{
		printf("%s",handlers[tmp][1]);
		tmp2=strlen(handlers[tmp][1]);
		while (tmp2 < handler_width)
		{
			putc(' ',stdout);
			tmp2++;
		}
		running_total += handler_width;
		tmp++;
	}
}

void append_prog(char *name)
{
	FILE *f;
	char buf[81];
	int x,y,eol;
	f = fopen(name,"r");
	eol = 0;
	for (y=0;y<25;y++)
	{
		if (eol != 2)
		{
			if (fgets(buf,81,f) == 0)
				eol = 2; // probably eof
			else
				eol=0;
		}
		for (x=0;x<80;x++)
		{
			if (eol)
				putc(' ',stdout);
			else if ((buf[x] != 10) && (buf[x] != 13))
				putc(buf[x],stdout);
			else
			{
				eol = 1;
				putc(' ',stdout);
			}
		}
	}
	fclose(f);
}

int main(int argc,char **argv)
{
	// Need to piece it all together
	// Work out approx size of items:
	// Code length=sum of segment lengths+(num. numbers needed*number_width)+handler width
	// Handler width = max handler string length
	// table size = no. entries*2
	// code block size = no. handlers*length
	// program offset = code length+table size+code block size
	int tmp,tmp2;
	handler_width = table_entries = tmp = 0;
	while (tmp < NUM_HANDLERS)
	{
		if (strlen(handlers[tmp][1]) > handler_width)
			handler_width = strlen(handlers[tmp][1]);
		table_entries += strlen(handlers[tmp][0]);
		tmp++;
	}
	table_entries++; // Stringmode entry
	fprintf(stderr,"handler_width=%d\ntable_entries=%d\nNUM_HANDLERS=%d\n",handler_width,table_entries,NUM_HANDLERS);
	codelen = numvars = 0;
	tmp = 0;
	while (codesegs[tmp])
	{
		tmp2=0;
		while (codesegs[tmp][tmp2])
		{
			if ((codesegs[tmp][tmp2] >= 'A') && (codesegs[tmp][tmp2] <= 'Z'))
				numvars++;
			else if ((codesegs[tmp][tmp2] != 'i') && (codesegs[tmp][tmp2] != 'h'))
				codelen++;
			tmp2++;
		}
		tmp++;
	}
	fprintf(stderr,"codelen=%d\nnumvars=%d\n",codelen,numvars);
	fprintf(stderr,"NUMBER_WIDTH=%d\nTABLE_SIZE=%d\nHANDLER_SIZE=%d\nTABLESETUP_SIZE=%d\nAPPROX_LEN=%d\n",NUMBER_WIDTH,TABLE_SIZE,HANDLER_SIZE,TABLESETUP_SIZE,APPROX_LEN);
	// Output code...
	tmp=0;
	running_total = 0;
	while (codesegs[tmp])
	{
		tmp2=0;
		while (codesegs[tmp][tmp2])
		{
			if (codesegs[tmp][tmp2] == 'X')
				output_x();
			else if (codesegs[tmp][tmp2] == 'Y')
				output_y();
			else if (codesegs[tmp][tmp2] == 'Z')
				output_z();
			else if (codesegs[tmp][tmp2] == 'W')
				output_w();
			else if (codesegs[tmp][tmp2] == 'h')
				output_h();
			else if (codesegs[tmp][tmp2] == 'i')
				output_i();
			else
			{
				running_total++;
				putc(codesegs[tmp][tmp2],stdout);
			}
			tmp2++;
		}
		tmp++;
	}
	output_table();
	output_handlers();
	if (running_total > APPROX_LEN)
	{
		fprintf(stderr,"\n*** Got length %d! overrun! ***\n",running_total);
		exit(1);
	}
	// Approx len should equal real len
	fprintf(stderr,"\n*** Result: %d chars ***\n",running_total);
	while (running_total<APPROX_LEN)
	{
		putc(' ',stdout);
		running_total++;
	}
	// Stick befunge program on as well
	if (argc == 2)
	{
		fprintf(stderr,"\n*** Appending program '%s'... ***\n",argv[1]);
		append_prog(argv[1]);
	}
	return 0;
}
